{$Q-,R-}
{***************************************************************
 *
 * Unit Name: SecDemoMain
 * Purpose  : Demonstration of StreamSec Security packages
 * Author   : Henrick Hellstrm, Copyright (c) 2001 StreamSec
 * History  : First version
 *
 ****************************************************************}

unit SecDemoMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, StreamSec, LarArith, DiffieHellman, SecUtils;

type
  TForm1 = class(TForm)
    GroupBox1: TGroupBox;
    edtPlainTextFile: TEdit;
    Label1: TLabel;
    Button1: TButton;
    Label2: TLabel;
    edtEncFile: TEdit;
    Button2: TButton;
    rgSymmetricEnc: TRadioGroup;
    GroupBox2: TGroupBox;
    Label3: TLabel;
    btnGenA: TButton;
    btnGenB: TButton;
    Label4: TLabel;
    btnGenSRP: TButton;
    Label5: TLabel;
    edtDecFile: TEdit;
    Button3: TButton;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    SaveDialog2: TSaveDialog;
    rgKeyExch: TRadioGroup;
    btnNegotiate: TButton;
    btnEncrypt: TButton;
    btnDecrypt: TButton;
    StreamSec1: TStreamSec;
    Button4: TButton;
    Button5: TButton;
    Button6: TButton;
    procedure btnGenAClick(Sender: TObject);
    procedure btnGenSRPClick(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure btnNegotiateClick(Sender: TObject);
    procedure btnEncryptClick(Sender: TObject);
    procedure btnDecryptClick(Sender: TObject);
    procedure StreamSec1Auth(Sender: TObject; ValidOutPos: Integer;
      Success: Boolean);
    procedure StreamSec1CodeBuf(Sender: TObject; InStream,
      OutStream: TStream; Done: Boolean);
    procedure Button6Click(Sender: TObject);
    procedure Button5Click(Sender: TObject);
    procedure Button4Click(Sender: TObject);
    procedure rgSymmetricEncClick(Sender: TObject);
  private
    { Keys for MTI/C0. Normally, the private keys are generated on the fly
      from the password, or stored somewhere in encrypted form.}
    PrivA: LarArith.TLargeInteger;
    SaltA: string;
    PrivB: LarArith.TLargeInteger;
    SaltB: string;
    CertA: TCert;        // TCert is declared in DiffieHellman
    CertB: TCert;
    CertBCreated: Boolean;
    CertACreated: Boolean;
    KeyNegotiated: Boolean;
    
    { Password verifier for SRP. The password verifier is kept by the server
      and should be stored somewhere safe.}
    SRPSaltA: string;
    PWVer: LarArith.TLargeInteger;
    { Symmetric encryption key. Normally generated on the fly and burned after
      it has been used. }
    Key: string;
  public
    // TCipherClass is declared in SecUtils
    function GetCipherClass: TCipherClass;
    procedure NegElGamal;
    procedure NegMTIC0;
    procedure NegSRP;
  end;

var
  Form1: TForm1;

implementation

uses
  PasswordMain, SteakRnd, Steak, Rijndael, TwoFish, Base64, LarBarret,
  SecDemoHexEditor;

{$R *.DFM}

procedure TForm1.btnGenAClick(Sender: TObject);
var
  PW, Salt: string;
  Priv: LarArith.TLargeInteger;
  Cert: TCert;
  IsA: Boolean;
begin
  IsA := Sender = btnGenA;

  { Firstly, prompt the user for a password: }
  // The PasswordDlg function is declared in PasswordMain
  if not PasswordDlg('Password for A''s MTI/C0 key',PW) then Exit;

  { Secondly, generate a random variable to diffuse the private key: }
  SetLength(Salt,16);
  // The SteakRandomBuf is declared in SteakRnd
  SteakRandomBuf(Salt[1],16);

  // Just to initialize the local variable:
  FillChar(Priv,SizeOf(Priv),0);

  { Thirdly, combine the password and the salt into a private key: }
  // The PasswordToPrivKey procedure is declared in DiffieHellman
  PasswordToPrivKey(PW,Salt,Priv[LarIntMaxSize - IntNorIndex],IntByteSize);

  { Fourthly, burn the password: }
  // The ProtectClear procedure is declared in SecUtils
  ProtectClear(PW[1],Length(PW));
              
  // Just to initialize the local variable:
  FillChar(Cert,SizeOf(Cert),0);

  { Fifthly, generate a public key: }
  // The PrivKeyToPublKey procedure is declared in DiffieHellman
  PrivKeyToPublKey(Priv[LarIntMaxSize - IntNorIndex],IntByteSize,Cert.PublKey,True);

  { Sixthly, build a certificate: }
  if IsA then
    Cert.Name := 'A'
  else
    Cert.Name := 'B';
  // The SignCertificate procedure is declared in DiffieHellman
  SignCertificate(Cert,nil,Priv);
  if IsA then begin
    PrivA := Priv;
    CertA := Cert;
    SaltA := Salt;
    CertACreated := True;
  end else begin
    PrivB := Priv;
    CertB := Cert;
    SaltB := Salt;
    CertBCreated := True;
  end;

  KeyNegotiated := False;

  { Lastly, burn all sensitive local variables: }
  ProtectClear(Priv,SizeOf(Priv));
  ProtectClear(Salt[1],Length(Salt));
end;

procedure TForm1.btnGenSRPClick(Sender: TObject);
var
  PW: string;
  ServerSRP: TServerSRP; // TServerSRP is declared in DiffieHellman
begin
  { Firstly, prompt the user for a password: }
  // The PasswordDlg function is declared in PasswordMain
  if not PasswordDlg('SRP password for A',PW) then Exit;
  try
    { The TServerSRP class will automatically build a verifier and a salt,
      if a non-empty password is passed to the constructor: }
    ServerSRP := TServerSRP.Create(PW);
    try
      SRPSaltA := ServerSRP.Salt;
      PWVer := ServerSRP.V;
    finally
      ServerSRP.Free;
    end;
  finally
    ProtectClear(PW[1],Length(PW));
  end;
end;

procedure TForm1.Button1Click(Sender: TObject);
begin
  if OpenDialog1.Execute then edtPlainTextFile.Text := OpenDialog1.FileName;
end;

procedure TForm1.Button2Click(Sender: TObject);
begin
  if SaveDialog1.Execute then edtEncFile.Text := SaveDialog1.FileName;
end;

procedure TForm1.Button3Click(Sender: TObject);
begin
  if SaveDialog2.Execute then edtDecFile.Text := SaveDialog2.FileName;
end;

function TForm1.GetCipherClass: TCipherClass;
begin
  case rgSymmetricEnc.ItemIndex of
    0: Result := TSteakCipher;
    1: Result := TTwoFish_PCFB;
    2: Result := TTwoFish_ABC;
    3: Result := TRijndael_PCFB;
    4: Result := TRijndael_ABC;
  else
    Result := nil;
  end;
end;

procedure TForm1.btnNegotiateClick(Sender: TObject);
begin
  KeyNegotiated := True;
  btnEncrypt.Enabled := True;
  btnDecrypt.Enabled := False;
  case rgKeyExch.ItemIndex of
    0: NegElGamal;
    1: NegMTIC0;
    2: NegSRP;
  end;
end;

procedure TForm1.NegElGamal;
var
  X: LarArith.TLargeInteger;
  K: LarArith.TLargeInteger;
  P: LarArith.TLargeInteger;
begin
  if not CertBCreated then
    raise Exception.Create('You must create a DL key for B first.');

  // Protocol actions A:
    // Initialize:
  FillChar(X,SizeOf(X),0);
  FillChar(K,SizeOf(K),0);
  FillChar(P,SizeOf(P),0);
    // Generate private key:
  SteakRandomBuf(X[LarIntMaxSize - IntNorIndex],IntByteSize);
    // Make sure it fits in the subgroup:
  while LargeCmpOffset(X,SysQ,0) >= 0 do LargeSub(X,SysQ);
    // Generate partial key data to send to B:
  LargeBarretPow4Mod(X,SysPrime,SysMicro,P);
    // Generate encryption key:
  LargeBarretExpMod(CertB.PublKey,X,SysPrime,SysMicro,K);
    // Set key:
  SetLength(Key,IntByteSize);
  Move(K[LarIntMaxSize - IntNorIndex],Key[1],IntByteSize);
                                             
  {
  // Protocol actions B:
  LargeBarretExpMod(K,PrivB,SysPrime,SysMicro,K);
    // Set key:
  SetLength(Key,IntByteSize);
  Move(K[LarIntMaxSize - IntNorIndex],Key[1],IntByteSize);
  }
  ShowMessage('Ready');
end;

procedure TForm1.NegMTIC0;
var
  A, B: TMTIC0;
  ZA, ZB: string;
begin         
  if not CertACreated then
    raise Exception.Create('You must create a DL key for A first.');
  if not CertBCreated then
    raise Exception.Create('You must create a DL key for B first.');
  A := TMTIC0.Create(CertB);
  try
    B := TMTIC0.Create(CertA);
    try
      ZA := A.Z;
      ZB := B.Z;
      B.SetZ(PrivB,ZA);
      A.SetZ(PrivA,ZB);
      // Addition, not part of the MTI/C0 protocol - verify success:
      A.M0 := B.M0;
      B.M1 := A.M1;
      if A.OK and B.OK then begin
        ShowMessage('Success');
        Key := A.K;
        {Key := B.K;}
      end else ShowMessage('Failure');
    finally
      B.Free;
    end;
  finally
    A.Free;
  end;
end;

procedure TForm1.NegSRP;
var
  A: TClientSRP;
  B: TServerSRP;
  PW: string;
  tmp: LarArith.TLargeInteger;
begin
  if SRPSaltA = '' then
    raise Exception.Create('You must create a SRP verifier first');
  if not PasswordDlg('Enter SRP password for A',PW) then Exit;
  try
    B := TServerSRP.Create('');
    try
      // A sends her username to B.
      // B retrieves A's verifier and salt from some secure storage:
      B.V := PWVer;
      B.Salt := SRPSaltA;
      // B sends A's salt to A:
      A := TClientSRP.Create(PW,SRPSaltA);
      try
        // A sends first message to B:
        tmp := A.A;
        B.SetA(tmp);
        // B sends second message to A:
        A.SetB(B.B,B.U);
        // A sends M1 to B:
        if B.VerifyM(A.M) then begin
          // B sends M2 to A:
          if A.VerifyM(B.M) then begin
            Key := A.K;
            ShowMessage('Success');
          end else
            ShowMessage('Failure');
        end else
          ShowMessage('Failure');
      finally
        A.Free;
      end;
    finally
      B.Free;
    end;
  finally
    ProtectClear(PW[1],Length(PW));
  end;
end;

procedure TForm1.btnEncryptClick(Sender: TObject);
var
  FSIn, FSOut: TFileStream;
  B: Byte;
begin
  if not KeyNegotiated then
    raise Exception.Create('You must negotiate a session key first');
  StreamSec1.CipherClass := GetCipherClass;
  StreamSec1.Lock(Key[1],Length(Key));
  StreamSec1.EncryptionMode := True;
  FSIn := TFileStream.Create(edtPlainTextFile.Text,fmOpenRead);
  try
    FSOut := TFileStream.Create(edtEncFile.Text,fmCreate);
    try
      FSIn.Position := 0;
      StreamSec1.CodeStream(FSIn,FSOut,True,False);
    finally
      FSOut.Free;
    end;
  finally
    FSIn.Free;
  end;
end;

procedure TForm1.btnDecryptClick(Sender: TObject);
var
  FSIn, FSOut: TFileStream;
  B: Byte;
begin      
  if not KeyNegotiated then
    raise Exception.Create('You must negotiate a session key first');
  StreamSec1.CipherClass := GetCipherClass;
  StreamSec1.Lock(Key[1],Length(Key));
  StreamSec1.EncryptionMode := False;
  FSIn := TFileStream.Create(edtEncFile.Text,fmOpenRead);
  try
    FSOut := TFileStream.Create(edtDecFile.Text,fmCreate);
    try
      StreamSec1.CodeStream(FSIn,FSOut,True,False);
    finally
      FSOut.Free;
    end;
  finally
    FSIn.Free;
  end;
  ShowMessage('Done');
end;

procedure TForm1.StreamSec1Auth(Sender: TObject; ValidOutPos: Integer;
  Success: Boolean);
begin
  if not Success then
    ShowMessage(Format('Failure at position %d',[ValidOutPos]));
end;

procedure TForm1.StreamSec1CodeBuf(Sender: TObject; InStream,
  OutStream: TStream; Done: Boolean);
begin
  if Done then ShowMessage('Done');
end;

procedure TForm1.Button6Click(Sender: TObject);
var
  FS: TFileStream;
  S: string;
begin
  FS := TFileStream.Create(edtDecFile.Text,fmOpenReadWrite or fmShareExclusive);
  try
    SetLength(S,FS.Size);
    FS.Read(Pointer(S)^,Length(S));
    Form2.OctetString := S;
    if Form2.ShowModal = mrOK then begin
      S := Form2.OctetString;
      FS.Size := 0;
      FS.Write(Pointer(S)^,Length(S));
    end;
  finally
    FS.Free;
  end;
end;

procedure TForm1.Button5Click(Sender: TObject);
var
  FS: TFileStream;
  S: string;
begin
  FS := TFileStream.Create(edtEncFile.Text,fmOpenReadWrite or fmShareExclusive);
  try
    SetLength(S,FS.Size);
    FS.Read(Pointer(S)^,Length(S));
    Form2.OctetString := S;
    if Form2.ShowModal = mrOK then begin
      S := Form2.OctetString;
      FS.Size := 0;
      FS.Write(Pointer(S)^,Length(S));
    end;
  finally
    FS.Free;
  end;
end;

procedure TForm1.Button4Click(Sender: TObject);
var
  FS: TFileStream;
  S: string;
begin
  FS := TFileStream.Create(edtPlainTextFile.Text,fmOpenReadWrite or fmShareExclusive);
  try
    SetLength(S,FS.Size);
    FS.Read(Pointer(S)^,Length(S));
    Form2.OctetString := S;
    if Form2.ShowModal = mrOK then begin
      S := Form2.OctetString;
      FS.Size := 0;
      FS.Write(Pointer(S)^,Length(S));
    end;
  finally
    FS.Free;
  end;
end;

procedure TForm1.rgSymmetricEncClick(Sender: TObject);
begin
  if rgSymmetricEnc.ItemIndex = 0 then
    StreamSec1.VectorSize := 8
  else
    StreamSec1.VectorSize := -1;
end;

end.
